/*
 * Copyright 2009 Arne Vajh�j.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <ssdef.h>
#include <rms.h>
#include <starlet.h>

#include "dk_vajhoej_vms_rms_nativeapi.h"

/*
 * 32 vs 64 bit pointer crap.
 */
#pragma pointer_size save
#pragma pointer_size 32
typedef char *char_ptr32;
typedef struct FAB *FAB_ptr32;
typedef struct RAB *RAB_ptr32;
#pragma pointer_size 64
typedef char *char_ptr64;
typedef struct FAB *FAB_ptr64;
typedef struct RAB *RAB_ptr64;
#pragma pointer_size restore

/*
 * Structure containing both FAB and RAB.
 * A pointer to this structures is used as "file pointer" and is passed between function calls.
 */
struct ctx
{
    struct FAB fab;
    struct RAB rab;
};

/*
 * Array to convert between the Java constants for search and the RMS constants.
 */
static int realcomp[] = { 0, RAB$M_KGE, RAB$M_KGT, RAB$M_KGE | RAB$M_REV, RAB$M_KGT | RAB$M_REV };

/*
 * Get record using specified key. The record is not locked.
 */
static int getany(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, void *k2, int ksize, jbyteArray buf)
{
    long int stat;
    int buflen;
    signed char *buf2;
    struct ctx *p;
    char _buf[32000]; // buffer that will work with 32 bit pointer
    p = (struct ctx *)ptr;
    buflen = (*cntx)->GetArrayLength(cntx, buf);
    if(buflen > 32000) buflen = 32000;
#if DEBUG
    printf("getany - buflen=%d\n", buflen);
#endif
    buf2 = (*cntx)->GetByteArrayElements(cntx, buf, NULL);
    p->rab.rab$l_kbf = (char_ptr32)(char *)k2;
    p->rab.rab$b_ksz = ksize;
    p->rab.rab$b_krf = ix;
    p->rab.rab$l_ubf = (char_ptr32)&_buf[0];
    p->rab.rab$w_usz = buflen;
    p->rab.rab$l_rop = RAB$M_NLK;
    stat = sys$get(&p->rab, 0, 0);
#if DEBUG
    printf("sys$get stat = %d\n", stat);
#endif
    memcpy(buf2, _buf, buflen);
    (*cntx)->ReleaseByteArrayElements(cntx, buf, buf2, 0);
    if(stat != RMS$_NORMAL) return 0;
    return buflen;
}

/*
 * Get record using specified key. Requires record to be locked.
 */
static int deleteany(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, void *k2, int ksize)
{
    long int stat;
    struct ctx *p;
    p = (struct ctx *)ptr;
    p->rab.rab$l_kbf = (char_ptr32)(char *)k2;
    p->rab.rab$b_ksz = ksize;
    p->rab.rab$b_krf = ix;
    stat = sys$delete(&p->rab, 0, 0);
#if DEBUG
    printf("sys$delete stat = %d\n", stat);
#endif
    if(stat != RMS$_NORMAL) return 0;
    return 1;
}

/*
 * Find record using specified key and criteria. The record will be locked if specified in opt.
 */
static int findany(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, void *k2, int ksize, int opt)
{
    long int stat;
    struct ctx *p;
    p = (struct ctx *)ptr;
    p->rab.rab$b_rac = RAB$C_KEY;
    p->rab.rab$l_kbf = (char_ptr32)(char *)k2;
    p->rab.rab$b_ksz = ksize;
    p->rab.rab$b_krf = ix;
    p->rab.rab$l_rop = opt;
    stat = sys$find(&p->rab, 0, 0);
#if DEBUG
    printf("sys$find stat = %d\n", stat);
#endif
    if(stat != RMS$_NORMAL) return 0;
    return 1;
}

/*
 * Open file.
 */
JNIEXPORT jlong JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_open(JNIEnv *cntx, jclass me, jstring fnm, jboolean ro)
{
    long int stat;
    const char *fnm2;
    char _fnm[1024];  // buffer that will work with 32 bit pointer
    struct ctx *p;
#if DEBUG
    printf("open - ro=%d\n", ro);
#endif
    fnm2 = (*cntx)->GetStringUTFChars(cntx, fnm, 0);
    strcpy(_fnm, fnm2);
    p = (struct ctx *)_malloc32(sizeof(struct ctx));
    p->fab = cc$rms_fab;
    p->fab.fab$l_fna = (char_ptr32)&_fnm[0];
    p->fab.fab$b_fns = strlen(fnm2);
    p->fab.fab$b_fac = ro ? FAB$M_GET : (FAB$M_GET | FAB$M_PUT | FAB$M_UPD | FAB$M_DEL);
    p->fab.fab$b_shr = FAB$M_SHRGET | FAB$M_SHRPUT | FAB$M_SHRUPD | FAB$M_SHRDEL;
    stat = sys$open(&p->fab, 0, 0);
#if DEBUG
    printf("sys$open stat = %d\n", stat);
#endif
    p->rab = cc$rms_rab;
    p->rab.rab$l_fab = (FAB_ptr32)&p->fab;
    p->rab.rab$b_rac = RAB$C_KEY;
    stat = sys$connect(&p->rab, 0, 0);
#if DEBUG
    printf("sys$connect stat = %d\n", stat);
#endif
    (*cntx)->ReleaseStringUTFChars(cntx, fnm, fnm2);
    if(stat != RMS$_NORMAL) return 0;
#if DEBUG
   printf("file open ptr=%ld\n", (long int)p);
#endif
    return (long int)p;
}

/*
 * Lock record using byte key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_lockByte(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jbyte k)
{
    char k2;
    k2 = k;
    return findany(cntx, me, ptr, ix, &k2, sizeof(k2), 0);
}

/*
 * Lock record using short key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_lockShort(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jshort k)
{
    short int k2;
    k2 = k;
    return findany(cntx, me, ptr, ix, &k2, sizeof(k2), 0);
}

/*
 * Lock record using int key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_lockInt(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jint k)
{
    long int k2;
    k2 = k;
    return findany(cntx, me, ptr, ix, &k2, sizeof(k2), 0);
}

/*
 * Lock record using long key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_lockLong(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jlong k)
{
    long long int k2;
    k2 = k;
    return findany(cntx, me, ptr, ix, &k2, sizeof(k2), 0);
}

/*
 * Lock record using string key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_lockString(JNIEnv * cntx, jclass me, jlong ptr, jbyte ix, jstring k)
{
    int res;
    const char *k2;
    k2 = (*cntx)->GetStringUTFChars(cntx, k, 0);
    res = findany(cntx, me, ptr, ix, (char *)k2, strlen(k2), 0);
    (*cntx)->ReleaseStringUTFChars(cntx, k, k2);
    return res;
}

/*
 * Get record using byte key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_getByte(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jbyte k, jbyteArray buf)
{
    char k2;
    k2 = k;
    return getany(cntx, me, ptr, ix, &k2, sizeof(k2), buf);
}

/*
 * Get record using short key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_getShort(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jshort k, jbyteArray buf)
{
    short int k2;
    k2 = k;
    return getany(cntx, me, ptr, ix, &k2, sizeof(k2), buf);
}

/*
 * Get record using int key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_getInt(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jint k, jbyteArray buf)
{
    long int k2;
    k2 = k;
    return getany(cntx, me, ptr, ix, &k2, sizeof(k2), buf);
}

/*
 * Get record using long key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_getLong(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jlong k, jbyteArray buf)
{
    long long int k2;
    k2 = k;
    return getany(cntx, me, ptr, ix, &k2, sizeof(k2), buf);
}

/*
 * Get record using String key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_getString(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jstring k, jbyteArray buf)
{
    int res;
    const char *k2;
    k2 = (*cntx)->GetStringUTFChars(cntx, k, 0);
    res = getany(cntx, me, ptr, ix, (char *)k2, strlen(k2), buf);
    (*cntx)->ReleaseStringUTFChars(cntx, k, k2);
    return res;
}

/*
 * Put record.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_put(JNIEnv *cntx, jclass me, jlong ptr, jbyteArray buf)
{
    long int stat;
    int buflen;
    signed char *buf2;
    struct ctx *p;
    signed char _buf[32000];  // buffer that will work with 32 bit pointer
    p = (struct ctx *)ptr;
    buflen = (*cntx)->GetArrayLength(cntx, buf);
    if(buflen > 32000) buflen = 32000;
#if DEBUG
    printf("put - buflen=%d\n", buflen);
#endif
    buf2 = (*cntx)->GetByteArrayElements(cntx, buf, NULL);
    memcpy(_buf, buf2, buflen);
    p->rab.rab$l_rbf = (char_ptr32)&_buf[0];
    p->rab.rab$w_rsz = buflen;
    stat = sys$put(&p->rab, 0, 0);
#if DEBUG
    printf("sys$put stat = %d\n", stat);
#endif
    (*cntx)->ReleaseByteArrayElements(cntx, buf, buf2, 0);
    if(stat != RMS$_NORMAL && stat != RMS$_OK_DUP) return 0;
    return buflen;
}

/*
 * Update record. Requires record to be locked.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_update(JNIEnv *cntx, jclass me, jlong ptr, jbyteArray buf)
{
    long int stat;
    int buflen;
    signed char *buf2;
    struct ctx *p;
    char _buf[32000]; // buffer that will work with 32 bit pointer
    p = (struct ctx *)ptr;
    buflen = (*cntx)->GetArrayLength(cntx, buf);
    if(buflen > 32000) buflen = 32000;
#if DEBUG
    printf("update - buflen=%d\n", buflen);
#endif
    buf2 = (*cntx)->GetByteArrayElements(cntx, buf, NULL);
    memcpy(_buf, buf2, buflen);
    p->rab.rab$l_rbf = (char_ptr32)&_buf[0];
    p->rab.rab$w_rsz = buflen;
    stat = sys$update(&p->rab, 0, 0);
#if DEBUG
    printf("sys$update stat = %d\n", stat);
#endif
    (*cntx)->ReleaseByteArrayElements(cntx, buf, buf2, 0);
    if(stat != RMS$_NORMAL) return 0;
    return buflen;
}

/*
 * Delete record using byte key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_deleteByte(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jbyte k)
{
    char k2;
    k2 = k;
    return deleteany(cntx, me, ptr, ix, &k2, sizeof(k2));
}

/*
 * Delete record using short key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_deleteShort(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jshort k )
{
    short int k2;
    k2 = k;
    return deleteany(cntx, me, ptr, ix, &k2, sizeof(k2));
}

/*
 * Delete record using int key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_deleteInt(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jint k)
{
    long int k2;
    k2 = k;
    return deleteany(cntx, me, ptr, ix, &k2, sizeof(k2));
}

/*
 * Delete record using long key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_deleteLong(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jlong k)
{
    long long int k2;
    k2 = k;
    return deleteany(cntx, me, ptr, ix, &k2, sizeof(k2));
}

/*
 * Delete record using string key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_deleteString(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jstring k)
{
    int res;
    const char *k2;
    k2 = (*cntx)->GetStringUTFChars(cntx, k, 0);
    res = deleteany(cntx, me, ptr, ix, (char *)k2, strlen(k2));
    (*cntx)->ReleaseStringUTFChars(cntx, k, k2);
    return res;
}

/*
 * Find record using byte key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_findByte(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jbyte k, jint comp)
{
    char k2;
    k2 = k;
    return findany(cntx, me, ptr, ix, &k2, sizeof(k2), realcomp[comp] | RAB$M_NLK);
}

/*
 * Find record using short key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_findShort(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jshort k, jint comp)
{
    short int k2;
    k2 = k;
    return findany(cntx, me, ptr, ix, &k2, sizeof(k2), realcomp[comp] | RAB$M_NLK);
}

/*
 * Find record using int key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_findInt(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jint k, jint comp)
{
    long int k2;
    k2 = k;
    return findany(cntx, me, ptr, ix, &k2, sizeof(k2), realcomp[comp] | RAB$M_NLK);
}

/*
 * Find record using long key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_findLong(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jlong k, jint comp)
{
    long long int k2;
    k2 = k;
    return findany(cntx, me, ptr, ix, &k2, sizeof(k2), realcomp[comp] | RAB$M_NLK);
}

/*
 * Find record using string key.
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_findString(JNIEnv *cntx, jclass me, jlong ptr, jbyte ix, jstring k, jint comp)
{
    int res;
    const char *k2;
    k2 = (*cntx)->GetStringUTFChars(cntx, k, 0);
    res = findany(cntx, me, ptr, ix, (char *)k2, strlen(k2), realcomp[comp] | RAB$M_NLK);
    (*cntx)->ReleaseStringUTFChars(cntx, k, k2);
    return res;
}

/*
 * Get record using current position..
 */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_get(JNIEnv * cntx, jclass me, jlong ptr, jbyteArray buf)
{
    long int stat;
    int buflen;
    signed char *buf2;
    struct ctx *p;
    char _buf[32000]; // buffer that will work with 32 bit pointer
    p = (struct ctx *)ptr;
    buflen = (*cntx)->GetArrayLength(cntx, buf);
    if(buflen > 32000) buflen = 32000;
#if DEBUG
    printf("get - buflen=%d\n", buflen);
#endif
    buf2 = (*cntx)->GetByteArrayElements(cntx, buf, NULL);
    p->rab.rab$b_rac = RAB$C_SEQ;
    p->rab.rab$l_ubf = (char_ptr32)&_buf[0];
    p->rab.rab$w_usz = buflen;
    p->rab.rab$l_rop = RAB$M_NLK;
    stat = sys$get(&p->rab, 0, 0);
#if DEBUG
    printf("sys$get stat = %d\n", stat);
#endif
    memcpy(buf2, _buf, buflen);
    (*cntx)->ReleaseByteArrayElements(cntx, buf, buf2, 0);
    if(stat != RMS$_NORMAL) return 0;
    return buflen;
}

/*
 * Close file.
 */
JNIEXPORT void JNICALL Java_dk_vajhoej_vms_rms_NativeAPI_close(JNIEnv *cntx, jclass me, jlong ptr)
{
    long stat;
    struct ctx *p;
    p = (struct ctx *)ptr;
    stat = sys$disconnect(&p->rab, 0, 0);
#if DEBUG
    printf("sys$disconnect stat = %d\n", stat);
#endif
    stat = sys$close(&p->fab, 0, 0);
#if DEBUG
    printf("sys$close stat = %d\n", stat);
#endif
    free(p);
#if DEBUG
   printf("file close ptr=%ld\n", (long int)p);
#endif
}
